from cfast_slic import SlicModel, is_supported_arch
from .base_slic import BaseSlic, LSC

if not is_supported_arch("x64/avx2"):
    raise ImportError(
        "fast_slic is not configured with avx2 support. "
        "Compile it again with flag USE_AVX2."
    )

class SlicAvx2(BaseSlic):
    arch_name = "x64/avx2"

class LSCAvx2(LSC):
    arch_name = "x64/avx2"



from cfast_slic import SlicModel

class BaseSlic(object):
    arch_name = "__TODO__"

    def __init__(self,
                 num_components=400,
                 slic_model=None,
                 compactness=10,
                 min_size_factor=0.25,
                 subsample_stride=3,
                 convert_to_lab=True,
                 preemptive=False,
                 preemptive_thres=0.05,
                 manhattan_spatial_dist=True,
                 debug_mode=False,
                 num_threads=-1):
        self.compactness = compactness
        self.subsample_stride = subsample_stride
        self.min_size_factor = min_size_factor
        self._slic_model = slic_model and slic_model.copy() or self.make_slic_model(num_components)
        self._last_assignment = None

        self.convert_to_lab = convert_to_lab
        self._slic_model.preemptive = preemptive
        self._slic_model.preemptive_thres = preemptive_thres
        self._slic_model.manhattan_spatial_dist = manhattan_spatial_dist
        self._slic_model.num_threads = num_threads
        self._slic_model.debug_mode = debug_mode

    @property
    def convert_to_lab(self):
        return self._slic_model.convert_to_lab

    @convert_to_lab.setter
    def convert_to_lab(self, v):
        self._slic_model.convert_to_lab = v

    @property
    def slic_model(self):
        return self._slic_model

    @property
    def last_assignment(self):
        return self._last_assignment

    def iterate(self, image, max_iter=10):
        if not self._slic_model.initialized:
            self._slic_model.initialize(image)
        assignment = self._slic_model.iterate(image, max_iter, self.compactness, self.min_size_factor, self.subsample_stride)
        self._last_assignment = assignment
        return assignment

    @property
    def num_components(self):
        return self._slic_model.num_components

    def make_slic_model(self, num_components):
        return SlicModel(num_components, self.arch_name)

class Slic(BaseSlic):
    arch_name = 'standard'

class SlicRealDist(BaseSlic):
    arch_name = 'standard'
    real_dist_type = 'standard'

    def make_slic_model(self, num_components):
        model = SlicModel(num_components, self.arch_name)
        model.real_dist = True
        model.real_dist_type = self.real_dist_type
        return model

class SlicRealDistL2(SlicRealDist):
    arch_name = 'standard'
    real_dist_type = 'l2'

class SlicRealDistNoQ(SlicRealDist):
    arch_name = 'standard'
    real_dist_type = 'noq'

    def __init__(self, *args, **kwargs):
        float_color = kwargs.pop("float_color", True)
        super(SlicRealDistNoQ, self).__init__(*args, **kwargs)
        self._slic_model.float_color = float_color

class LSC(SlicRealDist):
    arch_name = 'standard'
    real_dist_type = 'lsc'



# -*- coding: utf-8 -*-
#
#     Copyright (c) 2018 Anders Høst
#

from __future__ import print_function

import platform
import os
import ctypes
from ctypes import c_uint32, c_int, c_long, c_ulong, c_size_t, c_void_p, POINTER, CFUNCTYPE

# Posix x86_64:
# Three first call registers : RDI, RSI, RDX
# Volatile registers         : RAX, RCX, RDX, RSI, RDI, R8-11

# Windows x86_64:
# Three first call registers : RCX, RDX, R8
# Volatile registers         : RAX, RCX, RDX, R8-11

# cdecl 32 bit:
# Three first call registers : Stack (%esp)
# Volatile registers         : EAX, ECX, EDX

_POSIX_64_OPC = [
        0x53,                    # push   %rbx
        0x89, 0xf0,              # mov    %esi,%eax
        0x89, 0xd1,              # mov    %edx,%ecx
        0x0f, 0xa2,              # cpuid
        0x89, 0x07,              # mov    %eax,(%rdi)
        0x89, 0x5f, 0x04,        # mov    %ebx,0x4(%rdi)
        0x89, 0x4f, 0x08,        # mov    %ecx,0x8(%rdi)
        0x89, 0x57, 0x0c,        # mov    %edx,0xc(%rdi)
        0x5b,                    # pop    %rbx
        0xc3                     # retq
]

_WINDOWS_64_OPC = [
        0x53,                    # push   %rbx
        0x89, 0xd0,              # mov    %edx,%eax
        0x49, 0x89, 0xc9,        # mov    %rcx,%r9
        0x44, 0x89, 0xc1,        # mov    %r8d,%ecx
        0x0f, 0xa2,              # cpuid
        0x41, 0x89, 0x01,        # mov    %eax,(%r9)
        0x41, 0x89, 0x59, 0x04,  # mov    %ebx,0x4(%r9)
        0x41, 0x89, 0x49, 0x08,  # mov    %ecx,0x8(%r9)
        0x41, 0x89, 0x51, 0x0c,  # mov    %edx,0xc(%r9)
        0x5b,                    # pop    %rbx
        0xc3                     # retq
]

_CDECL_32_OPC = [
        0x53,                    # push   %ebx
        0x57,                    # push   %edi
        0x8b, 0x7c, 0x24, 0x0c,  # mov    0xc(%esp),%edi
        0x8b, 0x44, 0x24, 0x10,  # mov    0x10(%esp),%eax
        0x8b, 0x4c, 0x24, 0x14,  # mov    0x14(%esp),%ecx
        0x0f, 0xa2,              # cpuid
        0x89, 0x07,              # mov    %eax,(%edi)
        0x89, 0x5f, 0x04,        # mov    %ebx,0x4(%edi)
        0x89, 0x4f, 0x08,        # mov    %ecx,0x8(%edi)
        0x89, 0x57, 0x0c,        # mov    %edx,0xc(%edi)
        0x5f,                    # pop    %edi
        0x5b,                    # pop    %ebx
        0xc3                     # ret
]

is_windows = os.name == "nt"
is_64bit   = ctypes.sizeof(ctypes.c_voidp) == 8

class CPUID_struct(ctypes.Structure):
    _fields_ = [(r, c_uint32) for r in ("eax", "ebx", "ecx", "edx")]

class CPUID(object):
    def __init__(self):
        if platform.machine() not in ("AMD64", "x86_64", "x86", "i686"):
            raise SystemError("Only available for x86")
        
        if is_windows:
            if is_64bit:
                # VirtualAlloc seems to fail under some weird
                # circumstances when ctypes.windll.kernel32 is
                # used under 64 bit Python. CDLL fixes this.
                self.win = ctypes.CDLL("kernel32.dll")
                opc = _WINDOWS_64_OPC
            else:
                # Here ctypes.windll.kernel32 is needed to get the
                # right DLL. Otherwise it will fail when running
                # 32 bit Python on 64 bit Windows.
                self.win = ctypes.windll.kernel32
                opc = _CDECL_32_OPC
        else:
            opc = _POSIX_64_OPC if is_64bit else _CDECL_32_OPC

        size = len(opc)
        code = (ctypes.c_ubyte * size)(*opc)

        if is_windows:
            self.win.VirtualAlloc.restype = c_void_p
            self.win.VirtualAlloc.argtypes = [ctypes.c_void_p, ctypes.c_size_t, ctypes.c_ulong, ctypes.c_ulong]
            self.addr = self.win.VirtualAlloc(None, size, 0x1000, 0x40)
            if not self.addr:
                raise MemoryError("Could not allocate RWX memory")
        else:
            self.libc = ctypes.cdll.LoadLibrary(None)
            self.libc.valloc.restype = ctypes.c_void_p
            self.libc.valloc.argtypes = [ctypes.c_size_t]
            self.addr = self.libc.valloc(size)
            if not self.addr:
                raise MemoryError("Could not allocate memory")

            self.libc.mprotect.restype = c_int
            self.libc.mprotect.argtypes = [c_void_p, c_size_t, c_int]
            ret = self.libc.mprotect(self.addr, size, 1 | 2 | 4)
            if ret != 0:
                raise OSError("Failed to set RWX")


        ctypes.memmove(self.addr, code, size)

        func_type = CFUNCTYPE(None, POINTER(CPUID_struct), c_uint32, c_uint32)
        self.func_ptr = func_type(self.addr)

    def __call__(self, eax, ecx=0):
        struct = CPUID_struct()
        self.func_ptr(struct, eax, ecx)
        return struct.eax, struct.ebx, struct.ecx, struct.edx

    def __del__(self):
        if is_windows:
            self.win.VirtualFree.restype = c_long
            self.win.VirtualFree.argtypes = [c_void_p, c_size_t, c_ulong]
            self.win.VirtualFree(self.addr, 0, 0x8000)
        elif self.libc:
            # Seems to throw exception when the program ends and
            # libc is cleaned up before the object?
            self.libc.free.restype = None
            self.libc.free.argtypes = [c_void_p]
            self.libc.free(self.addr)

if __name__ == "__main__":
    def valid_inputs():
        cpuid = CPUID()
        for eax in (0x0, 0x80000000):
            highest, _, _, _ = cpuid(eax)
            while eax <= highest:
                regs = cpuid(eax)
                yield (eax, regs)
                eax += 1

    print(" ".join(x.ljust(8) for x in ("CPUID", "A", "B", "C", "D")).strip())
    for eax, regs in valid_inputs():
        print("%08x" % eax, " ".join("%08x" % reg for reg in regs))




import pytest
import os
import numpy as np
import ctypes as C

from fast_slic import Slic
from fast_slic.avx2 import SlicAvx2
from PIL import Image

data_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "data",
)

@pytest.fixture
def fish_image():
    with Image.open(os.path.join(data_dir, "fish.jpg")) as img:
        return np.array(img)

@pytest.fixture
def fish_image_result():
    with Image.open(os.path.join(data_dir, "fish_result.png")) as img:
        return np.array(img)

@pytest.fixture
def fish_image_avx2_result():
    with Image.open(os.path.join(data_dir, "fish_result.avx2.png")) as img:
        return np.array(img)

@pytest.fixture
def fish_image_01_result():
    with Image.open(os.path.join(data_dir, "fish_result.min_size_factor-0.1.png")) as img:
        return np.array(img)

@pytest.fixture
def fish_image_01_avx2_result():
    with Image.open(os.path.join(data_dir, "fish_result.min_size_factor-0.1.avx2.png")) as img:
        return np.array(img)


def test_slic_model_clusters():
    x = np.zeros([480, 640, 3], np.uint8)
    slic = Slic(num_components=100)
    slic.iterate(x)
    for i, cluster in enumerate(slic.slic_model.clusters):
        assert cluster['number'] == i
        assert isinstance(cluster, dict)
        assert len(cluster['yx']) == 2
        assert isinstance(cluster['yx'], tuple)
        assert len(cluster['color']) == 3
        assert isinstance(cluster['color'], tuple)
        assert isinstance(cluster['num_members'], int)



def test_slic_model_clusters_setter():
    x = np.zeros([480, 640, 3], np.uint8)
    slic = Slic(num_components=100)
    slic.iterate(x)
    orig_clusters = slic.slic_model.clusters
    slic.slic_model.clusters = orig_clusters[:10]
    assert len(slic.slic_model.clusters) == 10
    assert slic.slic_model.clusters == orig_clusters[:10]
    assert slic.slic_model.num_components == 10
    assert slic.num_components == 10


def test_slic(fish_image, fish_image_result, fish_image_avx2_result, fish_image_01_result, fish_image_01_avx2_result):
    assert (Slic(num_components=256, min_size_factor=0).iterate(fish_image) == fish_image_result).all()
    assert (SlicAvx2(num_components=256, min_size_factor=0).iterate(fish_image) == fish_image_avx2_result).all()
    assert (Slic(num_components=256, min_size_factor=0.1).iterate(fish_image) == fish_image_01_result).all()
    assert (SlicAvx2(num_components=256, min_size_factor=0.1).iterate(fish_image) == fish_image_01_avx2_result).all()





import numpy as np
import pytest

from fast_slic.crf import SimpleCRF

def test_crf_basic():
    crf = SimpleCRF(3, 100)
    assert crf.space_size == 300
    assert crf.first_time == -1
    assert crf.last_time == -1
    assert crf.num_frames == 0
    with pytest.raises(IndexError):
        crf.get_frame(10)
    assert crf.pop_frame() == -1


def test_crf_frame():
    crf = SimpleCRF(3, 100)
    frame = crf.push_frame()
    assert crf.num_frames == 1
    assert crf.first_time == frame.time
    assert crf.last_time == frame.time
    assert frame.space_size == 300
    assert frame.time == 0
    assert crf.get_frame(0).time == 0


def test_crf_frame_2():
    crf = SimpleCRF(3, 100)
    frame_1 = crf.push_frame()
    frame_2 = crf.push_frame()

    assert crf.num_frames == 2
    assert crf.first_time == frame_1.time
    assert crf.last_time == frame_2.time

    assert crf.pop_frame() == 0
    assert crf.first_time == crf.last_time == 1

def test_gc():
    crf = SimpleCRF(3, 100)
    frame = crf.push_frame()
    del crf
    
    import gc; gc.collect()
    # frame should be alive
    frame.unaries
    frame.get_inferred()

def test_unaries():
    crf = SimpleCRF(3, 3)
    frame = crf.push_frame()

    frame.set_unbiased()
    assert (frame.unaries == np.log(3)).all()

    # Active probability: 0.333 + (1 - 0.333) * 0.5 = 0.666667
    f = np.array([0, 1, 2], np.int32)
    frame.set_mask(f, 0.5)
    unaries = frame.unaries
    exp_unaries = -np.log(
        np.array(
            [[2/3., 1/6., 1/6.],
             [1/6., 2/3., 1/6.],
             [1/6., 1/6., 2/3.]]
        )
    )
    assert np.isclose(unaries, exp_unaries).all()


    # Unary: -log(probability)
    prob = np.array(
        [[0.7, 0.5, 0.1],
         [0.1, 0.3, 0.15],
         [0.2, 0.2, 0.75]],
        np.float32
    )
    frame.set_proba(prob)
    assert np.isclose(frame.unaries, -np.log(prob)).all()


def test_proba():
    crf = SimpleCRF(3, 3)
    frame = crf.push_frame()
    # Unary: -log(probability)
    prob = np.array(
        [[0.7, 0.5, 0.1],
         [0.1, 0.3, 0.15],
         [0.2, 0.2, 0.75]],
        np.float32
    )
    frame.set_proba(prob)

    assert np.isclose(frame.get_inferred(), 0).all()
    crf.initialize()
    assert np.isclose(frame.get_inferred(), prob).all()


def test_initial_inferred():
    crf = SimpleCRF(3, 3)
    frame = crf.push_frame()

    frame.set_unbiased()
    assert (frame.get_inferred() == 0).all()
    frame.reset_inferred()
    assert np.isclose(frame.get_inferred(), 1 / 3.).all()


def test_set_yxmrgb():
    crf = SimpleCRF(3, 3)
    frame = crf.push_frame()
    frame.set_yxmrgb(
        np.array(
            [
                [1,2,1,3,4,5],
                [6,7,2,8,9,10],
                [11,12,3, 13,14,15],
            ],
            np.int32,
        )
    )
    res = frame.get_yxmrgb()
    assert len(res) == 3
    assert res[0] == [1,2,1,3,4,5]
    assert res[1] == [6,7,2,8,9,10]
    assert res[2] == [11,12,3, 13,14,15]


def test_set_connectivity():
    crf = SimpleCRF(3, 3)
    frame = crf.push_frame()
    assert frame.get_connectivity() == [[], [], []]
    with pytest.raises(TypeError):
        frame.set_connectivity([None, None, None])
    frame.set_connectivity([[0, 1], [2], [0]])
    assert frame.get_connectivity() == [[0, 1], [2], [0]]


def test_spatial_energy():
    # weight * exp(-(rgb_i - rgb_j) ** 2 / (2 * srgb ** 2) + -(xy_i - xy_j) ** 2 / (2 * sxy ** 2))
    spatial_srgb = 3.5
    spatial_w = 1.9
    spatial_sxy = 2.4
    crf = SimpleCRF(3, 2)

    crf.spatial_srgb = spatial_srgb
    crf.spatial_w = spatial_w
    crf.spatial_sxy = spatial_sxy
    assert np.isclose(crf.spatial_srgb, spatial_srgb)
    assert np.isclose(crf.spatial_w, spatial_w)
    assert np.isclose(crf.spatial_sxy, spatial_sxy)

    frame = crf.push_frame()

    frame.set_yxmrgb(
        np.array(
            [
                [1, 1, 1, 1, 2, 6],
                [0, 0, 1, 4, 5, 3],
            ],
            np.int32,
        )
    )

    energy = spatial_w * np.exp(-((1 - 4) ** 2 + (2 - 5) ** 2 + (6 - 3) ** 2) / (2 * spatial_srgb ** 2) - ((1 - 0) ** 2 + (1 - 0)**2) / (2 * spatial_sxy ** 2))

    assert np.isclose(frame.spatial_pairwise_energy(0, 1), energy)
    assert np.isclose(frame.spatial_pairwise_energy(1, 0), energy)
    assert np.isclose(frame.spatial_pairwise_energy(0, 0), 0)
    assert np.isclose(frame.spatial_pairwise_energy(1, 1), 0)

    
def test_temporal_energy():
    # weight * exp(-(rgb_i(t) - rgb_i(T)) ** 2 / (2 * srgb ** 2)) 

    temporal_srgb = 3.5
    temporal_w = 1.9
    crf = SimpleCRF(3, 1)

    crf.temporal_srgb = temporal_srgb
    crf.temporal_w = temporal_w

    assert np.isclose(crf.temporal_srgb, temporal_srgb)
    assert np.isclose(crf.temporal_w , temporal_w)

    frame_1 = crf.push_frame()
    frame_2 = crf.push_frame()

    frame_1.set_yxmrgb(
        np.array(
            [
                [0, 0, 1, 1, 2, 6],
            ],
            np.int32,
        )
    )
    frame_2.set_yxmrgb(
        np.array(
            [
                [0, 0, 1, 4, 5, 3],
            ],
            np.int32,
        )
    )

    energy = temporal_w * np.exp(-(((1 - 4) ** 2 + (2 - 5) ** 2 + (6 - 3) ** 2) / (2 * temporal_srgb ** 2)))

    assert np.isclose(frame_1.temporal_pairwise_energy(0, frame_2), energy)
    assert np.isclose(frame_2.temporal_pairwise_energy(0, frame_1), energy)
    assert np.isclose(frame_1.temporal_pairwise_energy(0, frame_1), 0)
    assert np.isclose(frame_2.temporal_pairwise_energy(0, frame_2), 0)




from cfast_slic import is_supported_arch
from .base_slic import BaseSlic, LSC

if not is_supported_arch("arm/neon"):
    raise ImportError(
        "fast_slic is not configured with neon support. "
        "Compile it again with flag USE_NEON."
    )

class SlicNeon(BaseSlic):
    arch_name = "arm/neon"

class LSCNeon(LSC):
    arch_name = "arm/neon"



# cython: language_level=3, boundscheck=False
# distutils: language = c++

import numpy as np
cimport numpy as np
cimport cfast_slic as cs

from libc.stdlib cimport malloc, free
from libc.string cimport memset
from libc.stdint cimport uint32_t, int32_t
from libcpp.vector cimport vector
from cython.operator cimport dereference as deref, preincrement

cdef extern from "src/simple-crf.hpp":
    ctypedef int simple_crf_time_t

    ctypedef struct SimpleCRFParams:
        float spatial_w
        float temporal_w
        float spatial_srgb
        float temporal_srgb
        float spatial_sxy
        float spatial_smooth_w
        float spatial_smooth_sxy

    cdef cppclass CSimpleCRFFrame "SimpleCRFFrame":
        simple_crf_time_t time
        size_t num_classes
        size_t num_nodes

        void normalize()
        void set_clusters(const cs.Cluster* clusters)
        void get_clusters(cs.Cluster* clusters)
        void set_connectivity(const cs.Connectivity* conn)
        const vector[int]& connected_nodes(int node) except +

        void get_unary(float *unaries_out) const
        void set_unbiased()
        void set_mask(const int* classes, float confidence)
        void set_proba(const float* probas);
        void set_unary(const float* unaries);

        void get_inferred(float *out) const
        void reset_inferred() nogil

        size_t space_size()
        float calc_temporal_pairwise_energy(int node, const CSimpleCRFFrame& other) nogil
        float calc_spatial_pairwise_energy(int node_i, int node_j) nogil

    cdef cppclass CSimpleCRF "SimpleCRF":
        SimpleCRFParams params

        CSimpleCRF(size_t num_classes, size_t num_nodes) except +
        simple_crf_time_t get_first_time()
        simple_crf_time_t get_last_time()
        size_t get_num_frames()
        simple_crf_time_t pop_frame()
        CSimpleCRFFrame& get_frame(simple_crf_time_t time) except +
        CSimpleCRFFrame& push_frame() except +
        size_t space_size()
        void initialize() nogil
        void inference(size_t max_iter) nogil



cdef class SimpleCRFFrame:
    cdef CSimpleCRFFrame* _c_frame
    cdef readonly object parent_crf

    def __cinit__(self, parent_crf):
        self.parent_crf = parent_crf # Prevent unintended GC: because CRF object owns Frame, associated frame object would become dangling pointer if crf python object got collected. 


    @property
    def time(self):
        return self._c_frame.time

    @property
    def space_size(self):
        return self._c_frame.space_size()

    @property
    def unaries(self):
        cdef np.ndarray[np.float32_t,  ndim=2, mode='c'] unaries = self._fresh_buffer()
        self._c_frame.get_unary(&unaries[0, 0])
        return unaries

    def get_yxmrgb(self):
        result = []
        cdef cs.Cluster* clusters = <cs.Cluster*>malloc(sizeof(cs.Cluster) * self.num_nodes)
        try:
            self._c_frame.get_clusters(clusters)
            for i in range(self.num_nodes):
                result.append([
                    clusters[i].y,
                    clusters[i].x,
                    clusters[i].num_members,
                    clusters[i].r,
                    clusters[i].g,
                    clusters[i].b,
                ])
        finally:
            free(clusters)
        return result

    cpdef set_yxmrgb(self, int32_t[:,::1] yxmrgb):
        if self.num_nodes != yxmrgb.shape[0]:
            raise ValueError("Expected the first dimension of yxmrgb to equal to {}".format(self.num_nodes))
        if 6 != yxmrgb.shape[1]:
            raise ValueError("Expected the second dimension of yxmrgb to equal to 6")
        cdef int i, y, x, r, g, b
        cdef cs.Cluster *clusters = <cs.Cluster*>malloc(sizeof(cs.Cluster) * len(yxmrgb));
        try:
            for i in range(yxmrgb.shape[0]):
                clusters[i].y = yxmrgb[i, 0]
                clusters[i].x = yxmrgb[i, 1]
                clusters[i].num_members = yxmrgb[i, 2]
                clusters[i].r = yxmrgb[i, 3]
                clusters[i].g = yxmrgb[i, 4]
                clusters[i].b = yxmrgb[i, 5]
                clusters[i].number = i
            self._c_frame.set_clusters(clusters)
        finally:
            free(clusters)

    def get_connectivity(self):
        cdef size_t i
        cdef const vector[int]* nodes
        result = []
        for node in range(self.num_nodes):
            nodes = &self._c_frame.connected_nodes(node)
            result.append(deref(nodes))
        return result
    
    cpdef _set_slic_connectivity(self, cs.NodeConnectivity connectivity):
        self._c_frame.set_connectivity(connectivity._c_connectivity)

    def set_connectivity(self, connectivity):
        cdef cs.Connectivity c_conn
        cdef int i, k, num_neighbor
        cdef uint32_t neighbor
        cdef cs.Connectivity* c_connectivity
        if isinstance(connectivity, cs.NodeConnectivity):
            self._set_slic_connectivity(<cs.NodeConnectivity>connectivity)
            return 

        if len(connectivity) != self.num_nodes:
            raise ValueError("Expected len(connectivity) to be {}".format(self.num_nodes))

        c_conn.num_nodes = len(connectivity)
        c_conn.num_neighbors = NULL
        c_conn.neighbors = NULL
        try:
            c_conn.num_neighbors = <int*>malloc(sizeof(int) * len(connectivity))
            c_conn.neighbors = <uint32_t**>malloc(sizeof(uint32_t *) * len(connectivity))
            memset(c_conn.neighbors, 0, sizeof(uint32_t *) * len(connectivity))

            for i, neighbors in enumerate(connectivity):
                c_conn.num_neighbors[i] = len(neighbors)
                c_conn.neighbors[i] = <uint32_t*>malloc(sizeof(uint32_t) * len(neighbors))
                for k, neighbor in enumerate(neighbors):
                    c_conn.neighbors[i][k] = neighbor
            self._c_frame.set_connectivity(&c_conn)
        finally:
            if c_conn.neighbors is not NULL:
                for i in range(len(connectivity)):
                    if c_conn.neighbors[i] is not NULL:
                        free(c_conn.neighbors[i])
                free(c_conn.neighbors)

            if c_conn.num_neighbors is not NULL:
                free(c_conn.num_neighbors)



    @unaries.setter
    def unaries(self, float[:,::1] new_value):
        self._check_demension(new_value)
        self._c_frame.set_unary(&new_value[0, 0])

    @property
    def num_nodes(self):
        return self._c_frame.num_nodes

    @property
    def num_classes(self):
        return self._c_frame.num_classes

    def set_unbiased(self):
        self._c_frame.set_unbiased()

    def set_mask(self, int[::1] classes, float confidence):
        if <size_t>classes.shape[0] != self._c_frame.num_nodes:
            raise ValueError("The dimension of class array should match the number of nodes {}".format(self._c_frame.num_nodes))
        self._c_frame.set_mask(&classes[0], confidence)

    def set_proba(self, float[:, ::1] proba):
        self._check_demension(proba)
        self._c_frame.set_proba(&proba[0, 0])

    def get_inferred(self):
        cdef np.ndarray[np.float32_t,  ndim=2, mode='c'] proba = self._fresh_buffer()
        self._c_frame.get_inferred(&proba[0, 0])
        return proba

    def reset_inferred(self):
        self._c_frame.reset_inferred()

    def temporal_pairwise_energy(self, int node_i, SimpleCRFFrame other):
        cdef float result
        if not isinstance(other, SimpleCRFFrame):
            raise TypeError("not a crf frame")
        cdef CSimpleCRFFrame *other_frame = other._c_frame
        if <size_t>node_i >= self._c_frame.num_nodes:
            raise ValueError("node number is out of range")

        with nogil:
            result = self._c_frame.calc_temporal_pairwise_energy(node_i, deref(other_frame))
        return result

    def spatial_pairwise_energy(self, int node_i, int node_j):
        cdef float result
        if <size_t>node_i >= self._c_frame.num_nodes or <size_t>node_j >= self._c_frame.num_nodes:
            raise ValueError("node number is out of range")
        with nogil:
            result = self._c_frame.calc_spatial_pairwise_energy(node_i, node_j)
        return result

    cdef _check_demension(self, float[:, ::1] arr):
        if <size_t>arr.shape[0] != self._c_frame.num_classes:
            raise ValueError("The first dimension of array should match the number of classes {}".format(self._c_frame.num_classes))
        if <size_t>arr.shape[1] != self._c_frame.num_nodes:
            raise ValueError("The second dimension of array should match the number of nodes {}".format(self._c_frame.num_nodes))

    cdef np.ndarray[np.float32_t,  ndim=2, mode='c'] _fresh_buffer(self):
        return np.zeros(
            [self._c_frame.num_classes, self._c_frame.num_nodes],
            dtype=np.float32
        )



cdef class SimpleCRF:
    cdef CSimpleCRF* _c_crf
    def __cinit__(self, size_t num_classes, size_t num_nodes):
        self._c_crf = new CSimpleCRF(num_classes, num_nodes)

    @property
    def spatial_w(self):
        return self._c_crf.params.spatial_w

    @spatial_w.setter
    def spatial_w(self, float spatial_w):
        self._c_crf.params.spatial_w = spatial_w

    @property
    def spatial_srgb(self):
        return self._c_crf.params.spatial_srgb

    @spatial_srgb.setter
    def spatial_srgb(self, float spatial_srgb):
        self._c_crf.params.spatial_srgb = spatial_srgb

    @property
    def spatial_sxy(self):
        return self._c_crf.params.spatial_sxy

    @spatial_sxy.setter
    def spatial_sxy(self, float spatial_sxy):
        self._c_crf.params.spatial_sxy = spatial_sxy

    @property
    def temporal_w(self):
        return self._c_crf.params.temporal_w

    @temporal_w.setter
    def temporal_w(self, float temporal_w):
        self._c_crf.params.temporal_w = temporal_w

    @property
    def temporal_srgb(self):
        return self._c_crf.params.temporal_srgb

    @temporal_srgb.setter
    def temporal_srgb(self, float temporal_srgb):
        self._c_crf.params.temporal_srgb = temporal_srgb

    @property
    def spatial_smooth_w(self):
        return self._c_crf.params.spatial_smooth_w

    @spatial_smooth_w.setter
    def spatial_smooth_w(self, float spatial_smooth_w):
        self._c_crf.params.spatial_smooth_w = spatial_smooth_w

    @property
    def spatial_smooth_sxy(self):
        return self._c_crf.params.spatial_smooth_sxy

    @spatial_smooth_sxy.setter
    def spatial_smooth_sxy(self, float spatial_smooth_sxy):
        self._c_crf.params.spatial_smooth_sxy = spatial_smooth_sxy

    @property
    def first_time(self):
        return self._c_crf.get_first_time()

    @property
    def last_time(self):
        return self._c_crf.get_last_time()

    @property
    def num_frames(self):
        return self._c_crf.get_num_frames()

    @property
    def space_size(self):
        return self._c_crf.space_size()

    def get_frame(self, simple_crf_time_t time):
        cdef CSimpleCRFFrame* frame = &self._c_crf.get_frame(time)
        cdef SimpleCRFFrame result = SimpleCRFFrame(self)
        result._c_frame = frame
        return result

    cpdef push_slic_frame(self, slic, knn=None):
        frame = self.push_frame()
        frame.set_yxmrgb(slic.slic_model.to_yxmrgb())
        if knn is None:
            frame.set_connectivity(slic.slic_model.get_connectivity(slic.last_assignment))
        else:
            frame.set_connectivity(slic.slic_model.get_knn_connectivity(slic.last_assignment, knn))
        frame.set_unbiased()
        return frame

    cpdef push_frame(self):
        cdef CSimpleCRFFrame* frame = &self._c_crf.push_frame()
        cdef SimpleCRFFrame result = SimpleCRFFrame(self)
        result._c_frame = frame
        return result
      
    cpdef pop_frame(self):
        return self._c_crf.pop_frame()

    def initialize(self):
        with nogil:
            self._c_crf.initialize()

    def inference(self, size_t max_iter):
        with nogil:
            self._c_crf.inference(max_iter)

    def __dealloc__(self):
        del self._c_crf





# cython: language_level=3

from libc.stdint cimport uint8_t, uint32_t, uint16_t, int16_t
from libcpp cimport bool
from libcpp.string cimport string

cdef extern from "src/fast-slic-common.h":
    ctypedef struct Cluster:
        float y
        float x
        float r
        float g
        float b
        float a
        uint16_t number
        uint8_t is_active
        uint8_t is_updatable
        uint32_t num_members

    ctypedef struct Connectivity:
        int num_nodes;
        int *num_neighbors;
        uint32_t **neighbors;


cdef extern from "src/fast-slic.h":
    Connectivity* fast_slic_get_connectivity(int H, int W, int K, const uint16_t *assignment) nogil
    Connectivity* fast_slic_knn_connectivity(int H, int W, int K, const Cluster* clusters, int num_neighbors) nogil
    void fast_slic_free_connectivity(Connectivity* conn) nogil
    void fast_slic_get_mask_density(int H, int W, int K, const Cluster* clusters, const uint16_t* assignment, const uint8_t *mask, uint8_t *cluster_densities) nogil
    void fast_slic_cluster_density_to_mask(int H, int W, int K, const Cluster *clusters, const uint16_t* assignment, const uint8_t *cluster_densities, uint8_t *result) nogil

cdef extern from "src/context.h" namespace "fslic":
    cdef cppclass BaseContext[DistType]:
        int16_t subsample_stride_config
        int num_threads
        float compactness
        float min_size_factor
        bool convert_to_lab

        bool preemptive
        float preemptive_thres

        bool manhattan_spatial_dist
        bool debug_mode

        BaseContext(int H, int W, int K, const uint8_t* image, Cluster *clusters) except +
        void initialize_clusters() nogil
        void initialize_state() nogil
        bool parallelism_supported() nogil
        void iterate(uint16_t *assignment, int max_iter) nogil except +
        string get_timing_report();
        string get_recorder_report();

    cdef cppclass Context(BaseContext[uint16_t]):
        Context(int H, int W, int K, const uint8_t* image, Cluster *clusters) except +

    cdef cppclass ContextRealDist(BaseContext[float]):
        ContextRealDist(int H, int W, int K, const uint8_t* image, Cluster *clusters) except +

    cdef cppclass ContextRealDistL2(ContextRealDist):
        ContextRealDistL2(int H, int W, int K, const uint8_t* image, Cluster *clusters) except +

    cdef cppclass ContextRealDistNoQ(ContextRealDist):
        bool float_color
        ContextRealDistNoQ(int H, int W, int K, const uint8_t* image, Cluster *clusters) except +

    cdef cppclass ContextBuilder:
        ContextBuilder()
        ContextBuilder(const char* arch)
        const char** supported_archs()
        bool is_supported_arch()
        const char* get_arch()
        void set_arch(const char* arch)
        Context* build(int H, int W, int K, const uint8_t* image, Cluster *clusters)

cdef extern from "src/lsc.h" namespace "fslic":
    cdef cppclass ContextLSC(ContextRealDist):
        ContextLSC(int H, int W, int K, const uint8_t* image, Cluster *clusters) except +

    cdef cppclass ContextLSCBuilder:
        ContextLSCBuilder()
        ContextLSCBuilder(const char* arch)
        const char** supported_archs()
        bool is_supported_arch()
        const char* get_arch()
        void set_arch(const char* arch)
        ContextLSC* build(int H, int W, int K, const uint8_t* image, Cluster *clusters)


cdef extern from "src/cca.h" namespace "cca":
    cdef cppclass ConnectivityEnforcer:
        ConnectivityEnforcer(const uint16_t *labels, int H, int W, int K, int min_threshold)
        void execute(uint16_t *out)

cdef class NodeConnectivity:
    cdef Connectivity* _c_connectivity
    cpdef tolist(self)

    @staticmethod
    cdef create(Connectivity* conn)


cdef class SlicModel:
    cdef Cluster* _c_clusters
    cdef readonly int num_components
    cdef public int num_threads
    cdef public object initialized
    cdef public object arch_name
    cdef public object real_dist
    cdef public object real_dist_type
    cdef public object convert_to_lab
    cdef public object preemptive
    cdef public float preemptive_thres
    cdef public object manhattan_spatial_dist
    cdef public object debug_mode

    cdef public object float_color
    cdef public object last_timing_report
    cdef public object last_recorder_report

    cpdef void initialize(self, const uint8_t [:, :, ::1] image)
    cpdef iterate(self, const uint8_t [:, :, ::1] image, int max_iter, float compactness, float min_size_factor, uint8_t subsample_stride)
    cpdef get_connectivity(self, const int16_t[:,::1] assignments)
    cpdef get_knn_connectivity(self, const int16_t[:,::1] assignments, size_t num_neighbors)
    cpdef get_mask_density(self, const uint8_t[:, ::1] mask, const int16_t[:, ::1] assignments)
    cpdef broadcast_density_to_mask(self, const uint8_t[::1] densities, const int16_t[:, ::1] assignments);
    cdef _get_clusters(self)
    cdef _set_clusters(self, clusters)

cpdef is_supported_arch(arch_name)
cpdef get_supported_archs()

cpdef enforce_connectivity(const int16_t[:,::1] assignments, int min_threshold)



#!/usr/bin/env python

from setuptools import dist
dist.Distribution().fetch_build_eggs(['cython', 'numpy'])

import os
import platform
import numpy as np

from Cython.Build import cythonize
from setuptools import find_packages
from distutils.core import setup
from distutils.extension import Extension

# https://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils
import os
import multiprocessing
try:
    from concurrent.futures import ThreadPoolExecutor as Pool
except ImportError:
    from multiprocessing.pool import ThreadPool as LegacyPool

    # To ensure the with statement works. Required for some older 2.7.x releases
    class Pool(LegacyPool):
        def __enter__(self):
            return self

        def __exit__(self, *args):
            self.close()
            self.join()


def build_extensions(self):
    """Function to monkey-patch
    distutils.command.build_ext.build_ext.build_extensions

    """
    self.check_extensions_list(self.extensions)

    try:
        num_jobs = os.cpu_count()
    except AttributeError:
        num_jobs = multiprocessing.cpu_count()

    with Pool(num_jobs) as pool:
        results = pool.map(self.build_extension, self.extensions)
        list(results)

def compile(
    self, sources, output_dir=None, macros=None, include_dirs=None,
    debug=0, extra_preargs=None, extra_postargs=None, depends=None,
):
    """Function to monkey-patch distutils.ccompiler.CCompiler"""
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs
    )
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

    for obj in objects:
        try:
            src, ext = build[obj]
        except KeyError:
            continue
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

    # Return *all* object filenames, not just the ones we just built.
    return objects


from distutils.ccompiler import CCompiler
from distutils.command.build_ext import build_ext
build_ext.build_extensions = build_extensions


def _compile_and_check(c_content, compiler_args = []):
    import os, tempfile, subprocess, shutil
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    filename = r'test.c'
    with open(filename, 'w') as file:
        file.write(c_content)
    with open(os.devnull, 'w') as fnull:
        args = [os.environ.get("CC") or 'cc']
        args += compiler_args
        args.append(filename)
        result = subprocess.call(args, stdout=fnull, stderr=fnull)
    os.chdir(curdir)
    #clean up
    shutil.rmtree(tmpdir)
    return result == 0

def _check_openmp():
    import os, tempfile, subprocess, shutil

    # see http://openmp.org/wp/openmp-compilers/
    omp_test = \
        r"""
#include <omp.h>
#include <stdio.h>
    int main() {
#pragma omp parallel
    printf("Hello from thread %d, nthreads %d\n", omp_get_thread_num(), omp_get_num_threads());
    }
        """
    return _compile_and_check(omp_test, ['-fopenmp', '-lgomp'])

def _check_avx2():
    from cpuid.cpuid import CPUID
    try:
        # Invoke CPUID instruction with eax 0x7
        # ECX bit 5: AVX2 support
        # For more information, refer to https://en.wikipedia.org/wiki/CPUID
        input_eax = 0x7
        output_eax, output_ebx, output_ecx, output_edx = CPUID()(input_eax)
        bits = bin(output_ebx)[::-1]
        avx2_support = bits[5]
        return avx2_support == '1'
    except:
        return False

def _check_neon():
    neon_test = r"""
#include <arm_neon.h>
    int main() {
        const uint16x8_t constant = { 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF };
        const uint32x4_t vadded = vpaddlq_u16(constant);
        return 0;
    }
    """
    return _compile_and_check(neon_test, ["-mfpu=neon"])



extra_compile_args = []
extra_link_args = []
if platform.system() != 'Windows':
    extra_compile_args.append("-std=c++11")
    if _check_openmp():
        extra_compile_args.append('-fopenmp')
        extra_link_args.append('-lgomp')

    if _check_avx2():
        extra_compile_args.append("-DUSE_AVX2")
        extra_compile_args.append("-mavx2")
        extra_compile_args.append("-mfma")
    if platform.machine().lower().startswith("arm") and _check_neon():
        extra_compile_args.append("-DUSE_NEON")
        extra_compile_args.append("-mfpu=neon")

else:
    extra_compile_args.append("/openmp")
    if _check_avx2():
        extra_compile_args.append("/DUSE_AVX2")
        extra_compile_args.append("/arch:AVX2")
        extra_compile_args.append("/fp:fast")


cpp_sources = [
    "src/timer.cpp",
    "src/parallel.cpp",
    "src/fast-slic.cpp",
    "src/cca.cpp",
    "src/context.cpp",
    "src/context-impl.cpp",
    "src/lsc.cpp",
    "src/lsc-builder.cpp",
]


setup(
    name="fast-slic",
    version="0.4.0",
    description="Fast Slic Superpixel Implementation",
    author="Alchan Kim",
    author_email="a9413miky@gmail.com",
    setup_requires = ["cython", "numpy"],
    install_requires=["numpy"],
    python_requires=">=3.5",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    packages=find_packages(),
    ext_modules=cythonize(
        [
            Extension(
                "cfast_slic",
                include_dirs=[np.get_include()],
                sources=cpp_sources + ["cfast_slic.pyx"],
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                language="c++",
            ),
            Extension(
                "csimple_crf",
                include_dirs=[np.get_include()],
                sources=["src/simple-crf.cpp", "csimple_crf.pyx"],
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                language="c++",
            )
        ]
    )
)



# cython: language_level=3, boundscheck=False

cimport cfast_slic
cimport numpy as np

from cython.operator cimport dereference as deref

import numpy as np

from libc.stdint cimport uint8_t, int32_t, uint32_t, uint16_t, int16_t
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, memset


cdef class SlicModel:
    def __cinit__(self, int num_components, arch_name="standard", real_dist=False):
        arch_name = arch_name.encode("utf-8")
        cdef cfast_slic.ContextBuilder builder

        builder.set_arch(arch_name)
        if not builder.is_supported_arch():
            raise NotImplementedError("Unsupported arch " + repr(arch_name))

        if num_components >= 65534:
            raise ValueError("num_components cannot exceed 65534")
        elif num_components <= 0:
            raise ValueError("num_components should be a non-negative integer")

        self.num_components = num_components
        self.num_threads = -1
        self.arch_name = arch_name
        self.real_dist = real_dist
        self.real_dist_type = "standard"
        self.convert_to_lab = False
        self.float_color = True
        self.debug_mode = False

        self._c_clusters = <cfast_slic.Cluster *>malloc(sizeof(cfast_slic.Cluster) * num_components)
        memset(self._c_clusters, 0, sizeof(cfast_slic.Cluster) * num_components)
        self.initialized = False
        self.preemptive = False
        self.preemptive_thres = 0.05
        self.manhattan_spatial_dist = True

    def copy(self):
        result = SlicModel(self.num_components)
        memcpy(result._c_clusters, self._c_clusters, sizeof(cfast_slic.Cluster) * self.num_components)
        result.initialized = self.initialized
        return result

    cdef _get_clusters(self):
        cdef cfast_slic.Cluster* cluster
        cdef int i

        result = []
        for i in range(0, self.num_components):
            cluster = self._c_clusters + i
            result.append(
                dict(
                    number=cluster.number,
                    yx=(cluster.y, cluster.x),
                    color=(cluster.r,  cluster.g, cluster.b),
                    num_members=cluster.num_members,
                )
            )
        return result

    cdef _set_clusters(self, clusters):
        cdef uint8_t r, g, b
        cdef uint16_t x, y
        cdef uint32_t num_members
        cdef int num_new_clusters, i
        cdef cfast_slic.Cluster* new_clusters

        num_new_clusters = len(clusters)
        new_clusters = <cfast_slic.Cluster *>malloc(sizeof(cfast_slic.Cluster) * num_new_clusters)
        try:
            for i in range(num_new_clusters):
                dict_ = clusters[i]
                y, x = dict_['yx']
                r, g, b = dict_['color']
                num_members = dict_['num_members']

                new_clusters[i].number = i
                new_clusters[i].y = y
                new_clusters[i].x = x
                new_clusters[i].r = r
                new_clusters[i].g = g
                new_clusters[i].b = b
                new_clusters[i].num_members = num_members
        except:
            free(new_clusters)
            raise
        if self._c_clusters is not NULL:
            free(self._c_clusters)
        self._c_clusters = new_clusters
        self.num_components = num_new_clusters
        self.initialized = True

    def to_yxmrgb(self):
        cdef cfast_slic.Cluster* cluster
        cdef int i

        cdef np.ndarray[np.float, ndim=2, mode='c'] result = np.ndarray([self.num_components, 6], dtype=np.float)
        for i in range(0, self.num_components):
            cluster = self._c_clusters + i
            result[i, 0] = cluster.y
            result[i, 1] = cluster.x
            result[i, 2] = cluster.num_members
            result[i, 3] = cluster.r
            result[i, 4] = cluster.g
            result[i, 5] = cluster.b
        return result

    @property
    def clusters(self):
        return self._get_clusters()

    @clusters.setter
    def clusters(self, clusters):
        self._set_clusters(clusters)


    cpdef void initialize(self, const uint8_t [:, :, ::1] image):
        if image.shape[2] != 3:
            raise ValueError("nchan != 3")

        cdef cfast_slic.ContextBuilder builder
        cdef int H = image.shape[0]
        cdef int W = image.shape[1]
        cdef int K = self.num_components
        cdef cfast_slic.Cluster* c_clusters = self._c_clusters
        builder.set_arch(self.arch_name)

        cdef cfast_slic.Context *context = builder.build(
            H,
            W,
            K,
            &image[0, 0, 0],
            c_clusters,
        )
        try:
            with nogil:
                context.initialize_clusters()
        finally:
            del context
        self.initialized = True


    cpdef iterate(self, const uint8_t [:, :, ::1] image, int max_iter, float compactness, float min_size_factor, uint8_t subsample_stride):
        if not self.initialized:
            raise RuntimeError("Slic model is not initialized")
        if image.shape[2] != 3:
            raise ValueError("nchan != 3")

        cdef cfast_slic.ContextBuilder builder
        cdef int H = image.shape[0]
        cdef int W = image.shape[1]
        cdef int K = self.num_components
        cdef cfast_slic.Cluster* c_clusters = self._c_clusters
        builder.set_arch(self.arch_name)

        cdef np.ndarray[np.uint16_t, ndim=2, mode='c'] assignments = np.zeros([H, W], dtype=np.uint16)
        cdef cfast_slic.Context *context
        cdef cfast_slic.ContextRealDist *context_real_dist
        cdef cfast_slic.ContextRealDistNoQ *c_noq
        cdef ContextLSCBuilder lsc_builder
        lsc_builder.set_arch(self.arch_name)

        if not self.real_dist:
            context = builder.build(
                H,
                W,
                K,
                &image[0, 0, 0],
                c_clusters,
            )
            try:
                context.num_threads = self.num_threads
                context.compactness = compactness
                context.min_size_factor = min_size_factor
                context.subsample_stride_config = subsample_stride
                context.convert_to_lab = self.convert_to_lab
                context.preemptive = self.preemptive
                context.preemptive_thres = self.preemptive_thres
                context.manhattan_spatial_dist = self.manhattan_spatial_dist
                context.debug_mode = self.debug_mode
                with nogil:
                    context.initialize_state()
                    context.iterate(
                        <uint16_t *>&assignments[0, 0],
                        max_iter,
                    )
            finally:
                self.last_timing_report = context.get_timing_report().decode("utf-8")
                self.last_recorder_report = context.get_recorder_report()
                del context
        else:
            if self.real_dist_type == 'standard':
                context_real_dist = new cfast_slic.ContextRealDist(
                    H,
                    W,
                    K,
                    &image[0, 0, 0],
                    c_clusters,
                )
            elif self.real_dist_type == 'lsc':
                if not lsc_builder.is_supported_arch():
                    raise RuntimeError("LSC is not supported by arch " + repr(self.arch_name))
                context_real_dist = lsc_builder.build(
                    H,
                    W,
                    K,
                    &image[0, 0, 0],
                    c_clusters,
                )
            elif self.real_dist_type == 'l2':
                context_real_dist = new cfast_slic.ContextRealDistL2(
                    H,
                    W,
                    K,
                    &image[0, 0, 0],
                    c_clusters,
                )
            elif self.real_dist_type == 'noq':
                c_noq = new cfast_slic.ContextRealDistNoQ(
                    H,
                    W,
                    K,
                    &image[0, 0, 0],
                    c_clusters,
                )
                c_noq.float_color = self.float_color
                context_real_dist = c_noq
            else:
                raise RuntimeError("No such real_dist_type " + repr(self.real_dist_type))

            try:
                context_real_dist.num_threads = self.num_threads
                context_real_dist.compactness = compactness
                context_real_dist.min_size_factor = min_size_factor
                context_real_dist.subsample_stride_config = subsample_stride
                context_real_dist.convert_to_lab = self.convert_to_lab
                context_real_dist.preemptive = self.preemptive
                context_real_dist.preemptive_thres = self.preemptive_thres
                context_real_dist.manhattan_spatial_dist = self.manhattan_spatial_dist
                context_real_dist.debug_mode = self.debug_mode
                with nogil:
                    context_real_dist.initialize_state()
                    context_real_dist.iterate(
                        <uint16_t *>&assignments[0, 0],
                        max_iter,
                    )
            finally:
                self.last_timing_report = context_real_dist.get_timing_report().decode("utf-8")
                self.last_recorder_report = context_real_dist.get_recorder_report()
                del context_real_dist
        result = assignments.astype(np.int16)
        result[result == 0xFFFF] = -1
        return result

    cpdef get_connectivity(self, const int16_t[:,::1] assignments):
        cdef int H = assignments.shape[0]
        cdef int W = assignments.shape[1]
        cdef int K = self.num_components
        cdef int i, k;
        cdef uint32_t neighbor

        cdef Connectivity* conn;
        with nogil:
            conn = cfast_slic.fast_slic_get_connectivity(H, W, K, <const uint16_t *>&assignments[0, 0])
        return NodeConnectivity.create(conn)

    cpdef get_knn_connectivity(self, const int16_t[:,::1] assignments, size_t num_neighbors):
        cdef int H = assignments.shape[0]
        cdef int W = assignments.shape[1]
        cdef int K = self.num_components
        cdef cfast_slic.Cluster* c_clusters = self._c_clusters
        with nogil:
            conn = cfast_slic.fast_slic_knn_connectivity(H, W, K, c_clusters, num_neighbors)
        return NodeConnectivity.create(conn)

    cpdef get_mask_density(self, const uint8_t[:, ::1] mask, const int16_t[:, ::1] assignments):
        cdef int H = assignments.shape[0]
        cdef int W = assignments.shape[1]
        cdef int K = self.num_components
        cdef const cfast_slic.Cluster* _c_clusters = self._c_clusters

        if mask.shape[0] != H or mask.shape[1] != W:
            raise ValueError("The shape of mask does not match the one of assignments")
        cdef np.ndarray[np.uint8_t, ndim=1, mode='c'] densities = np.ndarray([K], dtype=np.uint8)

        with nogil:
            cfast_slic.fast_slic_get_mask_density(
                H,
                W,
                K,
                _c_clusters,
                <uint16_t *>&assignments[0, 0],
                &mask[0, 0],
                <uint8_t *>&densities[0],
            )
        return densities

    cpdef broadcast_density_to_mask(self, const uint8_t[::1] densities, const int16_t[:, ::1] assignments):
        cdef int H = assignments.shape[0]
        cdef int W = assignments.shape[1]
        cdef int K = self.num_components
        cdef const cfast_slic.Cluster* _c_clusters = self._c_clusters
        if densities.shape[0] != K:
            raise ValueError("The shape of densities should match the number of clusters")
        cdef np.ndarray[np.uint8_t, ndim=2, mode='c'] mask = np.ndarray([H, W], dtype=np.uint8)
        with nogil:
            cfast_slic.fast_slic_cluster_density_to_mask(
                H,
                W,
                K,
                _c_clusters,
                <uint16_t *>&assignments[0, 0],
                &densities[0],
                <uint8_t *>&mask[0, 0],
            )

        return mask

    def __dealloc__(self):
        if self._c_clusters is not NULL:
            free(self._c_clusters)

cdef class NodeConnectivity:
    @staticmethod
    cdef create(Connectivity* conn):
        cdef NodeConnectivity c = NodeConnectivity()
        c._c_connectivity = conn
        return c

    cpdef tolist(self):
        if self._c_connectivity is NULL:
            return []
        result = []
        for k in range(self._c_connectivity.num_nodes):
            k_neighbors = []
            for i in range(self._c_connectivity.num_neighbors[k]):
                neighbor = self._c_connectivity.neighbors[k][i]
                k_neighbors.append(neighbor)
            result.append(k_neighbors)
        return result

    def __dealloc__(self):
        if self._c_connectivity is not NULL:
            cfast_slic.fast_slic_free_connectivity(self._c_connectivity)


cpdef is_supported_arch(arch_name):
    cdef cfast_slic.ContextBuilder builder
    builder.set_arch(arch_name.encode("utf-8"))
    return builder.is_supported_arch()

cpdef get_supported_archs():
    cdef cfast_slic.ContextBuilder builder
    cdef const char **p = builder.supported_archs()
    cdef const char *s
    result = []
    while True:
        s = deref(p)
        if not s: break
        result.append(s.decode("utf-8"))
        p += 1
    return result

cpdef enforce_connectivity(const int16_t[:,::1] assignments, int min_threshold):
    cdef int H = assignments.shape[0]
    cdef int W = assignments.shape[1]
    cdef int K = 0
    cdef int i, j
    cdef uint16_t label
    for i in range(H):
        for j in range(W):
            label = <uint16_t>assignments[i, j]
            if label != 0xFFFF and K < label:
                K = label
    K += 1

    assignments
    cdef cfast_slic.ConnectivityEnforcer *enforcer = new cfast_slic.ConnectivityEnforcer(
        <uint16_t *>&assignments[0, 0],
        H,
        W,
        K,
        min_threshold,
    )
    try:
        enforcer.execute(<uint16_t *>&assignments[0, 0])
    finally:
        del enforcer
    return assignments
